; Lehmer Random Number Generators applied to MIDI generation

(require :asdf)
(asdf:load-system :smolmidi)    ; https://thrig.me/src/smolmidi.git
(block nil (setq *random-state* (make-random-state t)) (return))

(defmacro ldb32 (n) `(ldb (byte 32 0) ,n))

; "really horrible" -- Knuth
(defun RANDU (seed)
  (declare ((unsigned-byte 32) seed))
  (mod (ldb32 (* seed 65539)) 2147483648))
; NOTE the initial seed is supposed to be coprime to the modulus
(let ((seed (random 2147483648)))
  (format t "~&RANDU SEED ~a~&" seed)
  (defun coinflip ()
    (if (> (setf seed (RANDU seed)) 1073741824) 1 0)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; RHYTHM - so that the melody does not just plod along

(defun circular (items) (setf (cdr (last items)) items) items)

; earlier lists are more likely to be picked than later ones; durations
; must add up to 96 or some other means of switching between rhythms
; devised. one could also use a weighted random (walker alias method)
; here which might be simpler, or maybe a markov chain to move between
; different rhythmic patterns
(defparameter *beats*
  (make-array 4 :initial-contents (list
                                    (circular '(48 24 24))
                                    (circular '(48 48))
                                    (circular '(24 24 48))
                                    (circular '(24 24 24 24)))))

(defun getabeat (index &optional (stash *beats*))
  (pop (aref stash index)))

(let ((index 0) (sum 0))
  (defun nextbeat (fn)
    (let ((duration (getabeat index)))
      (incf sum duration)
      (when (> sum 96)
        (psetf sum 0 index (funcall fn))
        (setf duration (getabeat index))
        (incf sum duration))
      duration)))

; adapated from https://metacpan.org/pod/Music::Voss
(let ((n 0))
  (defun vosscoin (count &optional (base 2))
    (lambda ()
      (prog1
        (loop for k from 0 below count sum
              (if (zerop (mod n (expt base k))) (coinflip) 0))
        (incf n)))))

(let ((fn (vosscoin 3))) (defun duration () (nextbeat fn)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; MELODY - this uses a custom 8-bit L-PRNG

(defmacro ldb8 (n) `(ldb (byte 8 0) ,n))

(defun smolnum (modulus alpha iters submod)
  (declare ((unsigned-byte 8) modulus alpha submod))
  (loop for seed from 1 below modulus nconc
        (let ((state seed))
          (loop repeat iters collect
                (mod (setf state (mod (ldb8 (* state alpha)) modulus))
                     submod)))))

(defparameter *ps* (make-array 5 :initial-contents '(57 59 62 64 67)))
(defun getpitch (index &optional (ps *ps*)) (aref ps index))

(defun generate-midi ()
  (let ((numlist (smolnum 31 4 10 5))
        (track (smolmidi:new-track)))
    (loop for n in numlist do
          (let ((pitch (getpitch n)))
            (smolmidi:note-on  track 0          pitch 96)
            (smolmidi:note-off track (duration) pitch 0)))
    track))

(smolmidi:write-midi-file "out.midi" (list (generate-midi)))